<?php
/**
 * mwbModelCommand class file.
 * generates model class with relations from MySQL workbench mwb file.
 * @author Frank Liu <frank@orite.com>
 * @link http://www.orite.com.au/
 * @copyright Copyright &copy; 2008-2009 Orite Group Pty Ltd
 * @license http://www.yiiframework.com/license/
 * @version $Id: mwbModelCommand.php R1 2009-01-20$
 */

class mwbModelCommand extends CConsoleCommand
{
	/**
	 * @var string the template file for the model class.
	 * Defaults to null, meaning using 'framework/cli/views/shell/model/model_relations.php'.
	 */
	public $templateFile;

	public function getHelp()
	{
		return <<<EOD
Usage: wmbmodel <mwb-file>
This command generates model classes with relationships parsed from MySQL workbench mwb file.
 * mwb-file: required, mwb file.
EOD;
	}

	/**
	 * Execute the action.
	 * @param array command line parameters specific for this command
	 */
	public function run($args)
	{
		if(!isset($args[0]))
		{
			echo "Error: wmb file name is required.\n";
			echo $this->getHelp();
			return;
		}
		
		$templateFile=$this->templateFile===null?YII_PATH.'/cli/views/shell/model/model_relations.php':$this->templateFile;
		$basePath=Yii::getPathOfAlias('application.models');

		$tbls = $this->parsemwb($args[0]);
		foreach ($tbls as $tbl){
			$classFile=$basePath.DIRECTORY_SEPARATOR.$tbl->Name.'.php';
			$list=array(
				$tbl->Name.'.php'=>array(
					'source'=>$templateFile,
					'target'=>$classFile,
					'callback'=>array($this,'generateModel'),
					'params'=>array($tbl->Name,$tbl->dumpRelations()),
				),
			);
			$this->copyFiles($list);
			include_once($classFile);

			echo <<<EOD

The '{$tbl->Name}' class has been successfully created in the following file:
    $classFile

If you have a 'db' database connection, you can test it now with:
    \$model={$tbl->Name}::model()->find();
    print_r(\$model);

EOD;
		}
	}
	
	/**
	 * Execute the action.
	 * @param array command line parameters specific for this command
	 */

	private function parsemwb($file){
		$xml = simplexml_load_file("zip://".$file."#document.mwb.xml");

		//get Physical Model
		$pm = getNodeByAttr($xml->value[0], 'content-struct-name', 'workbench.physical.Model');
		$physicalModel = getNodeByAttr($pm->value[0], 'struct-name', 'workbench.physical.Model');

		//get mysql Catalog
		$mc = getNodeByAttr($physicalModel->value[0], 'struct-name', 'db.mysql.Catalog');

		//get mysql Schema
		$ms = getNodeByAttr($mc->value[0], 'content-struct-name', 'db.mysql.Schema');
		$mysqlSchema = getNodeByAttr($ms->value[0], 'struct-name', 'db.mysql.Schema');


		//get mysql Table
		$mt = getNodeByAttr($mysqlSchema->value[0], 'content-struct-name', 'db.mysql.Table');
		$tables = getNodeByAttr($mt->value[0], 'struct-name', 'db.mysql.Table');

		$mwbtbls = array();
		$mwbFKs = array();

		foreach($tables->value as &$tab){
			$tid = (string) $tab->attributes()->id;
			$mwbtbls[$tid] = new mwbTable($tab);

			$fks = getNodeByAttr($tab, 'content-struct-name', 'db.mysql.ForeignKey');
			foreach($fks->value[0]->children() as $fk) {
				$f = new mwbFK($fk, $tid);
				if (!empty($f->RefTid)) $mwbFKs[] = $f;
			}
		}

		// translate ids in fk to actual names
		foreach($mwbFKs as &$fk) {
			//get table names
			$fk->Table = &$mwbtbls[$fk->Tid];
			$fk->RefTable = &$mwbtbls[$fk->RefTid];

			/*get columns names
			$cn = array();
			foreach($fk->Col as $c){
				$cn[] = $fk->Table->Columns[$c];
			}
			$fk->Columns = implode(',',$cn);
			unset($fk->Col);

			$cn = array();
			foreach($fk->RefCol as $c){
				$cn[] = $fk->RefTable->Columns[$c];
			}
			$fk->RefColumns = implode(',',$cn);
			unset($fk->RefCol);
			*/

			$mwbtbls[$fk->Tid]->setRelation($fk);
			$mwbtbls[$fk->RefTid]->setRefRelation($fk);
		}

		foreach($mwbtbls as &$tbl){
			$tbl->isComposite();
		}

		return $mwbtbls;
	}

	public function generateModel($source,$params)
	{
		list($tableName,$tableRelations)=$params;
		$content=file_get_contents($source);
		$rules='';
		if(($db=Yii::app()->getDb())!==null)
		{
			$db->active=true;
			if(($table=$db->schema->getTable($tableName))!==null)
			{
				$required=array();
				$integers=array();
				$numerical=array();
				foreach($table->columns as $column)
				{
					if($column->isPrimaryKey && $table->sequenceName!==null)
						continue;
					if(!$column->allowNull)
						$required[]=$column->name;
					if($column->type==='integer')
						$integers[]=$column->name;
					else if($column->type==='double')
						$numerical[]=$column->name;
					else if($column->type==='string' && $column->size>0)
						$rules.="\n\t\t\tarray('{$column->name}','length','max'=>{$column->size}),";
				}
				if($required!==array())
					$rules.="\n\t\t\tarray('".implode(', ',$required)."', 'required'),";
				if($integers!==array())
					$rules.="\n\t\t\tarray('".implode(', ',$integers)."', 'numerical', 'integerOnly'=>true),";
				if($numerical!==array())
					$rules.="\n\t\t\tarray('".implode(', ',$numerical)."', 'numerical'),";

				//relations
				$relations = "";
				foreach($tableRelations as $key=>$tr){
						$relations .= "\n\t\t\t'".$key."'=>array(".$tr[1].",'".$tr[2]."','".$tr[3]."'),";
				}
			}
			else
				echo "Warning: the table '$tableName' does not exist in the database.\n";
		}
		else
			echo "Warning: you do not have a 'db' database connection as required by Active Record.\n";

		$tr=array(
			'{ClassName}'=>$tableName,
			'{TableName}'=>$tableName,
			'{Rules}'=>$rules,
			'{Relations}'=>$relations);

		return strtr($content,$tr);
	}
}

function getNodeByAttr(&$p, $a, $v, $d = true){
	$r = false;
	foreach($p->children() as $node){
		if($node[$a] == $v){
			$r->{$node->getName()}[] = $node;
		}
	}
	if($d) unset($p);

	return $r;
}

class mwbTable{
	var $Name, $Columns, $Relations;

	public function __construct($xml){
		$this->Columns = array();
		$this->Relations = array();
		$tn = getNodeByAttr($xml, 'key', 'name');
		$this->Name = (string) $tn->value[0];

		$cols = getNodeByAttr($xml, 'content-struct-name', 'db.mysql.Column');
		foreach($cols->value[0]->children() as $c) {
			$cid = (string) $c->attributes()->id;
			$cn = getNodeByAttr($c, 'key', 'name');
			$this->Columns[$cid] = (string) $cn->value[0];
		}
	}

	function setRelation($fk){
		$this->Relations[$fk->RefTable->Name] = array($fk->RefTable, 'self::BELONGS_TO', $fk->RefTable->Name , $fk->RefCol);
	}

	function setRefRelation($fk){
		$m = ($fk->Many == 0)? "self::HAS_ONE" : "self::HAS_MANY";
		$this->Relations[$fk->Table->Name] = array($fk->Table, $m, $fk->Table->Name, $fk->Col);
	}


	function isComposite(){
		if(sizeof($this->Relations) != 2) return false;
		$pk = array();
		foreach($this->Relations as $r){
			$pk = array_merge($pk, $r[0]->Relations[$this->Name][3]);
			if($r[1] != 'self::BELONGS_TO'){
				return false;
			}
		}
		if(sizeof(array_diff($pk, array_keys($this->Columns))) > 0) return false;
		list($r1, $r2) = array_keys($this->Relations);
		$fk = $this->Name."(".$this->getRelationColNames($this->Relations[$r1]).",".$this->getRelationColNames($this->Relations[$r2]).")";
		$this->Relations[$r1][0]->Relations[$r2] = array($this->Relations[$r2][0], "self::MANY_MANY", $r2, $fk);
		unset($this->Relations[$r1][0]->Relations[$this->Name]);
		$this->Relations[$r2][0]->Relations[$r1] = array($this->Relations[$r1][0], "self::MANY_MANY", $r1, $fk);
		unset($this->Relations[$r2][0]->Relations[$this->Name]);
		return true;
	}

	function dumpRelations(){
		foreach($this->Relations as &$r){
			$r[3] = $this->getRelationColNames($r);
			unset($r[0]);
		}
		return $this->Relations;
	}

	function getRelationColNames($r){
		if(is_array($r[3])){
			$cn = array();
			foreach($r[3] as $c){
				$cn[] = $r[0]->Columns[$c];
			}
			$r[3] = implode(',', $cn);
		}
		return $r[3];
	}
}

class mwbFK{
	var $Name, $Tid, $Col, $RefTid, $RefCol;
	var $Table, $RefTable, $Columns, $RefColumns, $Many;

	public function __construct($xml, $tid){
		$this->Tid = $tid;
		$this->Col = array();
		$this->RefCol = array();

		// name
		$nm = getNodeByAttr($xml, 'key', 'name', false);
		$this->Name = (string) $nm->value[0];

		// col ids
		$cs = getNodeByAttr($xml, 'key', 'columns', false);
		foreach($cs->value[0]->children() as $c) {
			$this->Col[] = (string) $c;
		}

		// many
		$m = getNodeByAttr($xml, 'key', 'many', false);
		$this->Many = (int) $m->value[0];

		// ref table id
		$rt = getNodeByAttr($xml, 'key', 'referencedTable', false);
		$this->RefTid = (string) $rt->link[0];

		// ref col ids
		$rcs = getNodeByAttr($xml, 'key', 'referencedColumns');
		foreach($rcs->value[0]->children() as $rc) {
			$this->RefCol[] = (string) $rc;
		}
	}
}
